@@ -14,20 +14,14 @@
 
 /* Thread-specific data */
 
+#include <errno.h>
 #include <stddef.h>
 #include <stdlib.h>
 #include "pthread.h"
 #include "internals.h"
 
-typedef void (*destr_function)(void *);
-
 /* Table of keys. */
 
-struct pthread_key_struct {
-  int in_use;                   /* already allocated? */
-  destr_function destr;         /* destruction routine */
-};
-
 static struct pthread_key_struct pthread_keys[PTHREAD_KEYS_MAX] =
   { { 0, NULL } };
 
@@ -55,7 +49,7 @@ int __pthread_key_create(pthread_key_t * key, destr_function destr)
   pthread_mutex_unlock(&pthread_keys_mutex);
   return EAGAIN;
 }
-weak_alias (__pthread_key_create, pthread_key_create)
+strong_alias (__pthread_key_create, pthread_key_create)
 
 /* Delete a key */
 
@@ -72,14 +66,15 @@ int pthread_key_delete(pthread_key_t key)
   }
   pthread_keys[key].in_use = 0;
   pthread_keys[key].destr = NULL;
-  /* Set the value of the key to NULL in all running threads, so that
-     if the key is reallocated later by pthread_key_create, its
+  /* Set the value of the key to NULL in all running threads, so
+     that if the key is reallocated later by pthread_key_create, its
      associated values will be NULL in all threads. */
   idx1st = key / PTHREAD_KEY_2NDLEVEL_SIZE;
   idx2nd = key % PTHREAD_KEY_2NDLEVEL_SIZE;
   th = self;
   do {
-    if (th->p_specific[idx1st] != NULL)
+    /* If the thread already is terminated don't modify the memory.  */
+    if (!th->p_terminated && th->p_specific[idx1st] != NULL)
       th->p_specific[idx1st][idx2nd] = NULL;
     th = th->p_nextlive;
   } while (th != self);
@@ -98,16 +93,16 @@ int __pthread_setspecific(pthread_key_t key, const void * pointer)
     return EINVAL;
   idx1st = key / PTHREAD_KEY_2NDLEVEL_SIZE;
   idx2nd = key % PTHREAD_KEY_2NDLEVEL_SIZE;
-  if (self->p_specific[idx1st] == NULL) {
-    self->p_specific[idx1st] =
-      calloc(PTHREAD_KEY_2NDLEVEL_SIZE, sizeof (void *));
-    if (self->p_specific[idx1st] == NULL)
+  if (THREAD_GETMEM_NC(self, p_specific[idx1st]) == NULL) {
+    void *newp = calloc(PTHREAD_KEY_2NDLEVEL_SIZE, sizeof (void *));
+    if (newp == NULL)
       return ENOMEM;
+    THREAD_SETMEM_NC(self, p_specific[idx1st], newp);
   }
-  self->p_specific[idx1st][idx2nd] = (void *) pointer;
+  THREAD_GETMEM_NC(self, p_specific[idx1st])[idx2nd] = (void *) pointer;
   return 0;
 }
-weak_alias (__pthread_setspecific, pthread_setspecific)
+strong_alias (__pthread_setspecific, pthread_setspecific)
 
 /* Get the value of a key */
 
@@ -120,11 +115,12 @@ void * __pthread_getspecific(pthread_key_t key)
     return NULL;
   idx1st = key / PTHREAD_KEY_2NDLEVEL_SIZE;
   idx2nd = key % PTHREAD_KEY_2NDLEVEL_SIZE;
-  if (self->p_specific[idx1st] == NULL || !pthread_keys[key].in_use)
+  if (THREAD_GETMEM_NC(self, p_specific[idx1st]) == NULL
+      || !pthread_keys[key].in_use)
     return NULL;
-  return self->p_specific[idx1st][idx2nd];
+  return THREAD_GETMEM_NC(self, p_specific[idx1st])[idx2nd];
 }
-weak_alias (__pthread_getspecific, pthread_getspecific)
+strong_alias (__pthread_getspecific, pthread_getspecific)
 
 /* Call the destruction routines on all keys */
 
@@ -140,35 +136,42 @@ void __pthread_destroy_specifics()
        round++) {
     found_nonzero = 0;
     for (i = 0; i < PTHREAD_KEY_1STLEVEL_SIZE; i++)
-      if (self->p_specific[i] != NULL)
+      if (THREAD_GETMEM_NC(self, p_specific[i]) != NULL)
         for (j = 0; j < PTHREAD_KEY_2NDLEVEL_SIZE; j++) {
           destr = pthread_keys[i * PTHREAD_KEY_2NDLEVEL_SIZE + j].destr;
-          data = self->p_specific[i][j];
+          data = THREAD_GETMEM_NC(self, p_specific[i])[j];
           if (destr != NULL && data != NULL) {
-            self->p_specific[i][j] = NULL;
+            THREAD_GETMEM_NC(self, p_specific[i])[j] = NULL;
             destr(data);
             found_nonzero = 1;
           }
         }
   }
   for (i = 0; i < PTHREAD_KEY_1STLEVEL_SIZE; i++) {
-    if (self->p_specific[i] != NULL) free(self->p_specific[i]);
+    if (THREAD_GETMEM_NC(self, p_specific[i]) != NULL)
+      free(THREAD_GETMEM_NC(self, p_specific[i]));
   }
 }
 
 /* Thread-specific data for libc. */
 
-int __libc_internal_tsd_set(enum __libc_tsd_key_t key, const void * pointer)
+static int
+libc_internal_tsd_set(enum __libc_tsd_key_t key, const void * pointer)
 {
   pthread_descr self = thread_self();
 
-  self->p_libc_specific[key] = (void *) pointer;
+  THREAD_SETMEM_NC(self, p_libc_specific[key], (void *) pointer);
   return 0;
 }
+int (*__libc_internal_tsd_set)(enum __libc_tsd_key_t key, const void * pointer)
+     = libc_internal_tsd_set;
 
-void * __libc_internal_tsd_get(enum __libc_tsd_key_t key)
+static void *
+libc_internal_tsd_get(enum __libc_tsd_key_t key)
 {
   pthread_descr self = thread_self();
 
-  return self->p_libc_specific[key];
+  return THREAD_GETMEM_NC(self, p_libc_specific[key]);
 }
+void * (*__libc_internal_tsd_get)(enum __libc_tsd_key_t key)
+     = libc_internal_tsd_get;
